"""

A VTK RenderWindowInteractor widget for wxPython.  Note that wxPython
comes with its own wxVTKRenderWindow in wxPython.lib.vtk.  Try both
and see which one works better for you.  

Find wxPython info at http://wxPython.org

Created by Prabhu Ramachandran, April 2002
Based on wxVTKRenderWindow.py

Fixes and updates by Charl P. Botha 2003-2005
"""

"""
Please see the example at the end of this file.

----------------------------------------
Creation:

 wxVTKRenderWindowInteractor(parent, ID, stereo=0, [wx keywords]):
 
 You should create a wx.PySimpleApp() or some other wx**App before
 creating the window.

Behaviour:

 Uses __getattr__ to make the wxVTKRenderWindowInteractor behave just
 like a vtkGenericRenderWindowInteractor.

----------------------------------------

"""

# import usual libraries
import math, os, sys

# wxPython and ParaView Server Manager
import wx, pvsm

# a few configuration items, see what works best on your system

# Use wxGLCanvas as base class instead of wx.Window.
# This is sometimes necessary under wxGTK or the image is blank.
# (in wxWindows 2.3.1 and earlier, the GLCanvas had scroll bars)
try:
    WX_USE_GL_CANVAS
except NameError:
    if wx.Platform == '__WXMSW__':
        WX_USE_GLCANVAS = 0
    else:
        WX_USE_GLCANVAS = 1

# Keep capturing mouse after mouse is dragged out of window
# (in wxGTK 2.3.2 there is a bug that keeps this from working,
# but it is only relevant in wxGTK if there are multiple windows)
try:
    WX_USE_X_CAPTURE
except NameError:
    if wx.Platform == '__WXMSW__': 
        WX_USE_X_CAPTURE = 1
    else:
        WX_USE_X_CAPTURE = 0

# end of configuration items


if WX_USE_GLCANVAS:
    import wx.glcanvas as gl
    baseClass = gl.GLCanvas
else:
    baseClass = wx.Window

class EventTimer(wx.Timer):
    def __init__(self, iren):
        wx.Timer.__init__(self)
        self.iren = iren
    def Notify(self):
        self.iren.TimerEvent()

class wxVTKRenderWindowInteractor(baseClass):
    """
    A wx.RenderWindow for wxPython.
    Use GetRenderWindow() to get the vtkRenderWindow.
    Create with the keyword stereo=1 in order to
    generate a stereo-capable window.
    """

    # class variable that can also be used to request instances that use
    # stereo; this is overridden by the stereo=1/0 parameter.  If you set
    # it to True, the NEXT instantiated object will attempt to allocate a
    # stereo visual.  E.g.:
    # wxVTKRenderWindowInteractor.USE_STEREO = True
    # myRWI = wxVTKRenderWindowInteractor(parent, -1)
    USE_STEREO = False
    
    def __init__(self, parent, ID, *args, **kw):

        # private attributes
        self.__RenderWhenDisabled = 0

        # First do special handling of some keywords:
        # stereo, position, size, width, height, style
        
        stereo = 0
        
        if kw.has_key('stereo'):
            if kw['stereo']:
                stereo = 1
            del kw['stereo']

        elif self.USE_STEREO:
            stereo = 1

        position = wx.DefaultPosition

        if kw.has_key('position'):
            position = kw['position']
            del kw['position']

        size = wx.DefaultSize

        if kw.has_key('size'):
            size = kw['size']
            del kw['size']
        
        if kw.has_key('width') and kw.has_key('height'):
            size = (kw['width'], kw['height'])
            del kw['width']
            del kw['height']

        # wx.WANTS_CHARS says to give us e.g. TAB
        # wx.NO_FULL_REPAINT_ON_RESIZE cuts down resize flicker under GTK
        style = wx.WANTS_CHARS | wx.NO_FULL_REPAINT_ON_RESIZE

        if kw.has_key('style'):
            style = style | kw['style']
            del kw['style']
            
        ren = None
        if kw.has_key('ren'):
            ren = kw['ren']
            del kw['ren']

        # the enclosing frame must be shown under GTK or the windows
        #  don't connect together properly
        l = []
        p = parent
        while p: # make a list of all parents
            l.append(p)
            p = p.GetParent()
        l.reverse() # sort list into descending order
        for p in l:
            p.Show(1)

        # code added by cpbotha to enable stereo correctly where the user
        # requests this; remember that the glXContext in this case is NOT
        # allocated by VTK, but by WX, hence all of this.
        if stereo and WX_USE_GLCANVAS:
            # initialize GLCanvas with correct attriblist for stereo
            attribList = [gl.WX_GL_RGBA, 
                          gl.WX_GL_MIN_RED, 1, gl.WX_GL_MIN_GREEN, 1,
                          gl.WX_GL_MIN_BLUE, 1, 
                          gl.WX_GL_DEPTH_SIZE, 1, gl.WX_GL_DOUBLEBUFFER,
                          gl.WX_GL_STEREO]
            try:
                baseClass.__init__(self, parent, ID, position, size, style, 
                                   attribList=attribList)
                
            except wx.PyAssertionError:
                # stereo visual couldn't be allocated, so we go back to default
                baseClass.__init__(self, parent, ID, position, size, style)
                # and make sure everyone knows about it
                stereo = 0

        else:
            baseClass.__init__(self, parent, ID, position, size, style)

        self._Ren = ren
        self._Ren.Register(None)
        
        self._Iren = ren.GetInteractor()
        self._Iren.AddObserver('CreateTimerEvent', self.CreateTimer)
        self._Iren.AddObserver('DestroyTimerEvent', self.DestroyTimer)
        self._Iren.AddObserver('RenderEvent', self.RenderCallback)

        self._Iren.GetRenderWindow().SetSize(size.width, size.height)
            
        if stereo:
            self._Iren.GetRenderWindow().StereoCapableWindowOn()
            self._Iren.GetRenderWindow().SetStereoTypeToCrystalEyes()

        self.__handle = None
        self._ActiveButton = 0

        self.BindEvents()

        # with this, we can make sure that the reparenting logic in
        # Render() isn't called before the first OnPaint() has
        # successfully been run (and set up the VTK/WX display links)
        self.__has_painted = False

    def BindEvents(self):
        # refresh window by doing a Render
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        # turn off background erase to reduce flicker
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda e: None)
        
        # Bind the events to the event converters
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnButtonDown)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnButtonDown)
        self.Bind(wx.EVT_MIDDLE_DOWN, self.OnButtonDown)
        self.Bind(wx.EVT_RIGHT_UP, self.OnButtonUp)
        self.Bind(wx.EVT_LEFT_UP, self.OnButtonUp)
        self.Bind(wx.EVT_MIDDLE_UP, self.OnButtonUp)
        self.Bind(wx.EVT_MOTION, self.OnMotion)

        # If we use wx.EVT_KEY_DOWN instead of wx.EVT_CHAR, capital versions
        # of all characters are always returned.  wx.EVT_CHAR also performs
        # other necessary keyboard-dependent translations.
        self.Bind(wx.EVT_CHAR, self.OnKeyDown)
        
        self.Bind(wx.EVT_SIZE, self.OnSize)
        

    def __getattr__(self, attr):        
        """Makes the object behave like a
        vtkGenericRenderWindowInteractor"""
        if attr == '__vtk__':
            return lambda t=self._Iren: t
        elif hasattr(self._Iren, attr):
            return getattr(self._Iren, attr)
        else:
            raise AttributeError, self.__class__.__name__ + \
                  " has no attribute named " + attr

    def CreateTimer(self, obj, evt):
        self._timer = EventTimer(self)
        self._timer.Start(10, True)

    def DestroyTimer(self, obj, evt):
        """The timer is a one shot timer so will expire automatically."""
        return 1
        
    def RenderCallback(self, obj, evt):
        self.Render()
    
    def GetDisplayId(self):
        """Function to get X11 Display ID from WX and return it in a format
        that can be used by VTK Python.

        We query the X11 Display with a new call that was added in wxPython
        2.6.0.1.  The call returns a SWIG object which we can query for the
        address and subsequently turn into an old-style SWIG-mangled string
        representation to pass to VTK.
        """
        
        d = None

        try:
            d = wx.GetXDisplay()
            
        except NameError:
            # wx.GetXDisplay was added by Robin Dunn in wxPython 2.6.0.1
            # if it's not available, we can't pass it.  In general, 
            # things will still work; on some setups, it'll break.
            pass
        
        else:
            # wx returns None on platforms where wx.GetXDisplay is not relevant
            if d:
                d = hex(d)
                # we now have 0xdeadbeef
                # VTK wants it as: _deadbeef_void_p (pre-SWIG-1.3 style)
                d = '_%s_%s' % (d[2:], 'void_p')

        return d
        
    def OnPaint(self,event):
        dc = wx.PaintDC(self)

        # make sure the RenderWindow is sized correctly
        self._Iren.GetRenderWindow().SetSize(self.GetSizeTuple())
        
        # Tell the RenderWindow to render inside the wxWindow.
        if not self.__handle:

            # on relevant platforms, set the X11 Display ID
            d = self.GetDisplayId()
            if d:
                self._Iren.GetRenderWindow().SetDisplayId(d)

            # store the handle
            self.__handle = self.GetHandle()
            # and give it to VTK
            self._Iren.GetRenderWindow().SetWindowInfo(str(self.__handle))

            # now that we've painted once, the Render() reparenting logic
            # is safe
            self.__has_painted = True
            
        self.Render()

    def OnSize(self,event):
        try:
            width, height = event.GetSize()
        except:
            width = event.GetSize().width
            height = event.GetSize().height
        self._Iren.SetSize(width, height)
        self._Iren.ConfigureEvent()
        # this will check for __handle
        self.Render()

        # event processing should continue
        event.Skip()

    def OnMotion(self, event):
        self._Iren.OnMove(event.GetX(), event.GetY())

        # event processing should continue
        event.Skip()

    def OnButtonDown(self,event):
        ctrl, shift = event.ControlDown(), event.ShiftDown()
                                            
        self._ActiveButton = 0
        if event.RightDown():
            self._Iren.OnRightPress(event.GetX(), event.GetY(), ctrl, shift)
            self._ActiveButton = 'Right'
        elif event.LeftDown():
            self._Iren.OnLeftPress(event.GetX(), event.GetY(), ctrl, shift)
            self._ActiveButton = 'Left'
        elif event.MiddleDown():
            self._Iren.OnMiddlePress(event.GetX(), event.GetY(), ctrl, shift)
            self._ActiveButton = 'Middle'

        # save the button and capture mouse until the button is released
        if self._ActiveButton and WX_USE_X_CAPTURE:
            self.CaptureMouse()

        # allow wx event processing to continue
        # on wxPython 2.6.0.1, omitting this will cause problems with
        # the initial focus, resulting in the wxVTKRWI ignoring keypresses
        # until we focus elsewhere and then refocus the wxVTKRWI frame
        event.Skip()

    def OnButtonUp(self,event):
        ctrl, shift = event.ControlDown(), event.ShiftDown()

        if self._ActiveButton == 'Right':
            self._Iren.OnRightRelease(event.GetX(), event.GetY(), ctrl, shift)
        elif self._ActiveButton == 'Left':
            self._Iren.OnLeftRelease(event.GetX(), event.GetY(), ctrl, shift)
        elif self._ActiveButton == 'Middle':
            self._Iren.OnMiddleRelease(event.GetX(), event.GetY(), ctrl, shift)

        # if the ActiveButton is realeased, then release mouse capture
        if self._ActiveButton and WX_USE_X_CAPTURE:
            self.ReleaseMouse()

        # event processing should continue
        event.Skip()
      
    def OnKeyDown(self,event):
        (x,y)= self._Iren.GetEventPosition()
        
        keycode = event.GetKeyCode()
        key = chr(0)
        if keycode < 256:
            key = chr(keycode)
        
        self._Iren.OnKeyPress(key, x, y)

        # event processing should continue
        event.Skip()
  
    def GetRenderWindow(self):
        return self._Iren.GetRenderWindow()

    def Render(self):
        RenderAllowed = 1
        
        if not self.__RenderWhenDisabled:
            # the user doesn't want us to render when the toplevel frame
            # is disabled - first find the top level parent
            topParent = wx.GetTopLevelParent(self)
            if topParent:
                # if it exists, check whether it's enabled
                # if it's not enabeld, RenderAllowed will be false
                RenderAllowed = topParent.IsEnabled()
            
        if RenderAllowed:
            if self.__handle and self.__handle == self.GetHandle():
                #self._Iren.GetRenderWindow().Render()
                self._Ren.StillRender()

            elif self.GetHandle() and self.__has_painted:
                # this means the user has reparented us; let's adapt to the
                # new situation by doing the WindowRemap dance
                self._Iren.GetRenderWindow().SetNextWindowInfo(
                    str(self.GetHandle()))

                # make sure the DisplayId is also set correctly
                d = self.GetDisplayId()
                if d:
                    self._Iren.GetRenderWindow().SetDisplayId(d)
                
                # do the actual remap with the new parent information
                self._Iren.GetRenderWindow().WindowRemap()

                # store the new situation
                self.__handle = self.GetHandle()
                
                # vtkPVRenderView plays here with timers and callbacks
                # and I'm not sure which render function should be called
                #self._Iren.GetRenderWindow().Render()
                self._Ren.StillRender()

    def SetRenderWhenDisabled(self, newValue):
        """Change value of __RenderWhenDisabled ivar.

        If __RenderWhenDisabled is false (the default), this widget will not
        call Render() on the RenderWindow if the top level frame (i.e. the
        containing frame) has been disabled.

        This prevents recursive rendering during wx.SafeYield() calls.
        wx.SafeYield() can be called during the ProgressMethod() callback of
        a VTK object to have progress bars and other GUI elements updated -
        it does this by disabling all windows (disallowing user-input to
        prevent re-entrancy of code) and then handling all outstanding
        GUI events.
        
        However, this often triggers an OnPaint() method for wxVTKRWIs,
        resulting in a Render(), resulting in Update() being called whilst
        still in progress.
        """
        self.__RenderWhenDisabled = bool(newValue)  

#-----------------------------------------------------------------------------
class MainApp(wx.App):
  def OnInit(self):
    self.frame = MainFrame(title='Parallel wxVTK prototype', size=(800,600))
    self.frame.Show(True)
    return True
    
  def OnExit(self):
    pvsm.ProxyManager.UnRegisterProxies()
    return 0

class MainFrame( wx.Frame ):
  def __init__( self, title, pos=wx.DefaultPosition, size=wx.DefaultSize, style=wx.DEFAULT_FRAME_STYLE ):
    wx.Frame.__init__( self, None, wx.ID_ANY, title, pos, size, style )
    
    self.wxVTK = wxVTKRenderWindowInteractor(self, wx.ID_ANY, ren=self.SetupSM())
    
    self.Bind( wx.EVT_CLOSE, self.OnCloseWindow )

  def OnCloseWindow(self, event):
    # UnRegister should be in wxVTKRWI class...
    self.wxVTK._Ren.UnRegister(None) 
    self.Destroy()
    
  def SetupSM(self):
    # Simple pipeline example
    pvsm.Application.ParseConfigurationFiles()
    
    cone1 = pvsm.ProxyManager.NewProxy('sources', 'ConeSource')
    pvsm.ProxyManager.RegisterProxy('sources', 'cone1', cone1)
    cone1.UnRegister(None)
    cone1.UpdateVTKObjects()
    
    cone2 = pvsm.ProxyManager.NewProxy('sources', 'ConeSource')
    pvsm.ProxyManager.RegisterProxy('sources', 'cone2', cone2)
    cone2.UnRegister(None)
    cone2.GetProperty('Resolution').SetElements1(30)
    cone2.UpdateVTKObjects()
    
    display1 = pvsm.ProxyManager.NewProxy('displays', 'LODDisplay')
    pvsm.ProxyManager.RegisterProxy('displays', 'display1', display1)
    display1.UnRegister(None)
    display1.GetProperty('Input').AddProxy(cone1)
    display1.GetProperty('Representation').SetElements1(2)
    display1.UpdateVTKObjects()
    
    display2 = pvsm.ProxyManager.NewProxy('displays', 'LODDisplay')
    pvsm.ProxyManager.RegisterProxy('displays', 'display2', display2)
    display2.UnRegister(None)
    display2.GetProperty('Input').AddProxy(cone2)
    display2.GetProperty('Representation').SetElements1(2)
    display2.GetProperty('Position').SetElements3(0.7, 0.5, 0)
    display2.UpdateVTKObjects()
    
    axes1 = pvsm.ProxyManager.NewProxy('axes', 'Axes')
    pvsm.ProxyManager.RegisterProxy('axes', 'axes1', axes1)
    axes1.UnRegister(None)
    axes1.UpdateVTKObjects()
      
    ren = pvsm.ProxyManager.NewProxy('rendermodules', 'LODRenderModule')
    pvsm.ProxyManager.RegisterProxy('rendermodules', 'ren', ren)
    ren.UnRegister(None)
    ren.GetProperty('Displays').AddProxy(display1)
    ren.GetProperty('Displays').AddProxy(axes1)
    ren.GetProperty('Displays').AddProxy(display2)
    ren.GetProperty('CameraPosition').SetElements3(0, 0, 3.20361)
    ren.UpdateVTKObjects()
    
    return ren

if __name__ == "__main__":
  app = MainApp()
  app.MainLoop()
